/*
 * Programming in Java, Jacobs University, 2015
 *
 * This file demonstrates how to implement a simple chat server. The
 * ChatDispatcher maintains information about the threads dealing with
 * the clients and it provides a central queue over which it receives
 * messages and a seperate queue for each client output thread to
 * forward messages to. Note that it is crucial for the hashmap to be
 * a thread-safe concurrent map.
 */

import java.util.Map;
import java.util.HashMap;
import java.net.Socket;
import java.io.IOException;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.concurrent.ConcurrentHashMap;

public class ChatDispatcher
    implements Runnable
{
    private static final int QUEUE_LIMIT = 42;
    
    private Map<Socket, BlockingQueue<String>> map
	= new ConcurrentHashMap<Socket, BlockingQueue<String>>();
    private BlockingQueue<String> queue
	= new LinkedBlockingDeque<String>(QUEUE_LIMIT);

    public void addClient(Socket s)
	throws IOException
    {
	BlockingQueue<String> q = new LinkedBlockingDeque<String>(QUEUE_LIMIT);
	
	Runnable ri = new ServeChatClientInput(s, this, queue);
	(new Thread(ri)).start();
	
	Runnable ro = new ServeChatClientOutput(s, this, q);
	(new Thread(ro)).start();
	map.put(s, q);
	q.offer(">> welcome to the chat room <<");

	broadcast(">> client " + ChatServer.getClientName(s) + " joined <<");
    }

    public void removeClient(Socket s)
    {
	map.remove(s);
	broadcast(">> client " + ChatServer.getClientName(s) + " left <<");
    }

    public void broadcast(String s)
    {
	for (Map.Entry<Socket, BlockingQueue<String>> entry: map.entrySet()) {
	    entry.getValue().offer(s);
	}
    }

    public void run()
    {
	while (! Thread.currentThread().isInterrupted()) {
	    String s = null;
	    try {
		s = queue.take();
	    } catch (InterruptedException e) {
		// reassert the interrupt to maintain the interrupt state
		Thread.currentThread().interrupt();
	    }
	    if (s != null) {
		broadcast(s);
	    }
	}
    }
}
